/* cut-down version of mgquery */


#include <stdio.h>
#include <string.h>
/* #include <io.h> */
#include <fcntl.h>

#ifdef __cplusplus
extern "C" {
#endif
  
#include "sysfuncs.h"
  
#include "messages.h"
#include "memlib.h"
  
#include "invf.h"
#include "text.h"
#include "mg.h"
#include "lists.h"
#include "backend.h"
#include "environment.h"
#include "globals.h"
#include "mg_errors.h"
#include "commands.h"
#include "text_get.h"
#include "term_lists.h"
#include "local_strings.h"
  
#ifdef __cplusplus
}
#endif


#include "mgq.h"

/* globals needed by some vague part of mg... */
FILE *OutFile = NULL, *InFile = NULL;
int OutPipe = 0, InPipe = 0;
int Quitting = 0;

static query_data *qd;
char db_loaded_name[256] = "";

#if defined(PARADOCNUM) ||  defined(NZDL)
int *Paragraph = NULL;
int Documents = 0;

int GetDocNumFromParaNum(int paranum)
{
  int low = 1, high = Documents;
  int mid = (low+high)/2;

  while ((mid = (low+high)/2) >=1 && mid <= Documents)
    {
      if (paranum > Paragraph[mid])
        low = mid+1;
      else if (paranum <= Paragraph[mid-1])
        high = mid-1;
      else
        return mid;
    }
  FatalError(1, "Bad paragraph number.\n");
  return 0;
}
#endif


/*****************************************************************************/

static void MGQError(char *emsg)
{
  fprintf(stderr,"Fatal error: %s\n", emsg);
  exit(1);
}

static int ProcessDocs (query_data * qd, int num, int skip, int howmany, int verbatim,
			enum result_kinds kind, FILE * Output,
			int (*sender)(char *,int,int,float,void *), void *ptr)
{
  int max_buf = 0, output_failure = 0;
  int DocCount = 0;
  int need_text = (kind == result_docs);
  
  for (;;) {
    if (skip <= 0) break;
    skip--;
    if (!NextDoc(qd)) return 0;
  }
  
  if (need_text)
    {
      max_buf = atoi (GetDefEnv ("buffer", "1048576"));
    }
  
  do
    {
      u_char *UDoc = NULL;
      unsigned long ULen;
      
      if (need_text)
	{
	  /* load the compressed text */
	  if (LoadCompressedText (qd, max_buf))
	    {
	      MGQError("Unable to load compressed text(memory?).");
	    }
	  /* uncompress the loaded text */
	  UDoc = GetDocText (qd, &ULen);
	  if (UDoc == NULL)
	    MGQError("UDoc is unexpectedly NULL");
	}
      
      if (UDoc != NULL  ||  kind == result_docnums)
	{
	  int docnum = GetDocNum(qd);
#if defined(PARADOCNUM) ||  defined(NZDL)
	  if (qd->id->ifh.InvfLevel == 3) docnum = GetDocNumFromParaNum(docnum);
#endif
	  switch (kind)
	    {
	    case result_docnums:
	      if (sender != NULL)
		output_failure = (*sender)("",0,docnum,GetDocWeight(qd),ptr);
	      break;
	    case result_docs:
	      if (sender != NULL)
		output_failure = (*sender)(UDoc,ULen,docnum,GetDocWeight(qd),ptr);
	      break;
	    default:
	      break;
	    }
	}
      DocCount++;
      
    }
  while (NextDoc (qd) && output_failure == 0 && --howmany > 0);
  
  if (need_text)
    {
      FreeTextBuffer (qd);
    }
  
  return (DocCount);
}


void
send_query_term_freqs(QueryTermList *qtl,
		      int (*sender)(char *,int,int,float,void *), void *ptr)
{
  int i;  char *word;
  for (i = 0; i < qtl->num; i++)
    if (sender != NULL) {
      word = word2str(qtl->QTE[i].Term);
      (* sender)(word, strlen(word), qtl->QTE[i].Count, (float)0.0, ptr);
    }
}


void
send_terms (TermList * qtl, int (*sender)(char *,int,int,float,void *), void *ptr)
{
  int i = 0;
  char *word;

  if (sender == NULL) return;
  for (i = 0; i < qtl->num; i++)
    {
      word = word2str(qtl->TE[i].Word);
      (* sender)(word, strlen(word), qtl->TE[i].Count, (float)0.0, ptr);
    }
}


/* MoreDocs () */
/* Displays all documents in list  DocList. */
/* Documents are fetched, then decompressed and displayed according to the */
/* format implied in  FormString(). */

static void
MoreDocs (query_data * qd, enum result_kinds kind, 
	  int skip, int howmany,
	  int (*sender)(char *,int,int,float,void *), void *ptr)
{
  qd->num_of_ans = qd->DL->num;
  switch (kind) {
  case result_docs:
  case result_docnums:
    if (qd->num_of_ans > 0)
      ProcessDocs (qd, qd->DL->num, skip, howmany,
		   BooleanEnv (GetEnv ("verbatim"), 1),
		   kind, NULL, sender, ptr);
    break;
  case result_termfreqs:
    send_query_term_freqs(qd->QTL, sender, ptr);
    break;
  case result_terms:
    send_terms(qd->TL, sender, ptr);
    break;
  }
}


int mgq_init(char *mgdir, char *gensuffix, char *textsuffix)
{
  /* Initialise the environment with default values */
  InitEnv ();
  
  SetEnv("mgdir",mgdir,NULL);
  SetEnv("mgname",gensuffix,NULL);
  SetEnv("textname",textsuffix,NULL);
  
  PushEnv ();
  
  qd = InitQuerySystem (GetDefEnv ("mgdir", ""),
			GetDefEnv ("mgname", ""),
			GetDefEnv ("textname", NULL),
			NULL);
  
  if (qd == NULL) return 0;  else return 1;
}

int mgq_ask(char *line)
{
  char QueryType;  char OutputType;  char *LinePtr;
  
  ResetFileStats (qd);
  qd->max_mem_in_use = qd->mem_in_use = 0;
  qd->tot_hops_taken += qd->hops_taken;
  qd->tot_num_of_ptrs += qd->num_of_ptrs;
  qd->tot_num_of_accum += qd->num_of_accum;
  qd->tot_num_of_terms += qd->num_of_terms;
  qd->tot_num_of_ans += qd->num_of_ans;
  qd->tot_text_idx_lookups += qd->text_idx_lookups;
  qd->hops_taken = qd->num_of_ptrs = 0;
  qd->num_of_accum = qd->num_of_ans = qd->num_of_terms = 0;
  qd->text_idx_lookups = 0;
  
  LinePtr = ProcessCommands (line, qd);
  if (CommandsErrorStr) {
    fprintf (stderr, "%s\n", CommandsErrorStr);
    return 0;
  }
  if (*LinePtr == '\0') return 1;
  
  FreeQueryDocs (qd);
  
  QueryType = get_query_type ();
  OutputType = get_output_type ();
  /* No point in hiliting words on a docnum query */
  if (OutputType == OUTPUT_HILITE && QueryType == QUERY_DOCNUMS)
    OutputType = OUTPUT_TEXT;
  
  switch (QueryType)
    {
    case QUERY_BOOLEAN:
      {
	char *maxdocs;
	BooleanQueryInfo bqi;
	maxdocs = GetDefEnv ("maxdocs", "all");
	bqi.MaxDocsToRetrieve = strcmp (maxdocs, "all") ? atoi (maxdocs) : -1;
        if (qd->sd->sdh.indexed)
	  BooleanQuery (qd, line, &bqi, (BooleanEnv (GetEnv ("casefold"), 0) |
					 (BooleanEnv (GetEnv ("stem"), 0) << 1)));
	else
	  BooleanQuery (qd, line, &bqi, qd->sd->sdh.stem_method);
	/*	if (qd->sd->sdh.indexed) BooleanQuery (qd, line, &bqi, 3);
	else BooleanQuery (qd, line, &bqi, qd->sd->sdh.stem_method); */
	break;
      }
    case QUERY_APPROX:
    case QUERY_RANKED:
      {
	char *maxdocs;
	char *maxterms;
	char *maxaccum;
	RankedQueryInfo rqi;
	maxdocs = GetDefEnv ("maxdocs", "all");
	maxterms = GetDefEnv ("max_terms", "all");
	maxaccum = GetDefEnv ("max_accumulators", "all");
	rqi.Sort = BooleanEnv (GetEnv ("sorted_terms"), 0);
	rqi.QueryFreqs = BooleanEnv (GetEnv ("qfreq"), 1);
	rqi.Exact = QueryType == QUERY_RANKED;
	rqi.MaxDocsToRetrieve = strcmp (maxdocs, "all") ? atoi (maxdocs) : -1;
	rqi.MaxTerms = strcmp (maxterms, "all") ? atoi (maxterms) : -1;
	rqi.MaxParasToRetrieve = rqi.MaxDocsToRetrieve;
	if (qd->id->ifh.InvfLevel == 3 && GetEnv ("maxparas"))
	  rqi.MaxParasToRetrieve = atoi (GetEnv ("maxparas"));
	rqi.AccumMethod = toupper (*GetDefEnv ("accumulator_method", "A"));
	rqi.MaxAccums = strcmp (maxaccum, "all") ? atoi (maxaccum) : -1;
	rqi.HashTblSize = IntEnv (GetEnv ("hash_tbl_size"), 1000);
	rqi.StopAtMaxAccum = BooleanEnv (GetEnv ("stop_at_max_accum"), 0);
	rqi.skip_dump = GetEnv ("skip_dump");
	RankedQuery (qd, line, &rqi);
	break;
      }
    case QUERY_DOCNUMS:
      {
	DocnumsQuery (qd, line);
	break;
      }
    }

  return 1;
}

int mgq_numdocs(void)
{
  if (qd->DL) return qd->DL->num;
  else return 0;
}

int mgq_numterms(void)
{
  if (qd->QTL) return qd->QTL->num;
  else return 0;
}

int mgq_results(enum result_kinds kind,int skip,int howmany, int (*sender)(char *, int, int, float, void *), void *ptr)
{
  if (qd->DL) {
    qd->doc_pos = 0;
    MoreDocs(qd, kind, skip, howmany, sender, ptr);
  }
  return 0;
}


void mgq_finish(void)
{
  FinishQuerySystem(qd);
  UninitEnv ();
}

int load_database(char *mgdir, char *gensuffix, char *textsuffix)
{
  if ((db_loaded_name[0]) != '\0' && (strcmp(gensuffix,db_loaded_name) != 0)) {
    mgq_finish();
    db_loaded_name[0] = '\0';
  }
  if ((db_loaded_name[0] == '\0') && mgq_init(mgdir, gensuffix, textsuffix)) {
    strcpy(db_loaded_name, gensuffix);
  }
  return (db_loaded_name[0] != '\0');
}
